import itertools as it
from dataclasses import dataclass
from typing import Any, Iterator, TypeAlias

from ..base import BaseSolver

Node: TypeAlias = tuple[int, int]
Edge: TypeAlias = tuple[Node, Node]


@dataclass(frozen=True)
class Region:
    value: str
    cells: set[tuple[int, int]]
    edges: set[Edge]
    sides: list[tuple[Edge, ...]]


def extract_region(grid: list[str], cell: tuple[int, int]):
    n_rows, n_cols = len(grid), len(grid[0])
    row, col = cell

    value = grid[row][col]

    cells: set[tuple[int, int]] = set()
    edges: set[Edge] = set()
    sides: list[tuple[Edge, ...]] = []

    queue: list[tuple[int, int]] = [(row, col)]
    while queue:
        row, col = queue.pop(0)

        if (row, col) in cells:
            continue

        cells.add((row, col))

        for ur, uc in (
            (row - 1, col),
            (row, col + 1),
            (row + 1, col),
            (row, col - 1),
        ):
            if 0 <= ur < n_rows and 0 <= uc < n_cols and grid[ur][uc] == value:
                queue.append((ur, uc))
                continue

            if ((row, col), (ur, uc)) in edges:
                continue

            if col == uc:
                mid, max = col, n_cols

                def get(v: int):
                    return (row, v, ur, v)
            else:
                mid, max = row, n_rows

                def get(v: int):
                    return (v, col, v, uc)

            side: tuple[Edge, ...] = ((((row, col), (ur, uc))),)
            for rng in (range(mid - 1, -1, -1), range(mid, max)):
                for r2, c2, ur2, uc2 in map(get, rng):
                    if grid[r2][c2] != value or (
                        0 <= ur2 < n_rows
                        and 0 <= uc2 < n_cols
                        and grid[r2][c2] == grid[ur2][uc2]
                    ):
                        break
                    side += ((((r2, c2), (ur2, uc2))),)

            sides.append(side)
            edges = edges.union(side)

    return Region(value=value, cells=cells, edges=edges, sides=sides)


class Solver(BaseSolver):
    def solve(self, input: str) -> Iterator[Any]:
        grid = input.splitlines()

        regions: list[Region] = []
        to_visit: set[tuple[int, int]] = set(
            it.product(range(len(grid)), range(len(grid[0])))
        )

        while to_visit:
            region = extract_region(grid, next(iter(to_visit)))

            self.logger.info(
                f"region with {region.value}: "
                f"{len(region.cells)} * {len(region.edges)} = {len(region.cells) * len(region.edges)}, "
                f"{len(region.cells)} * {len(region.sides)} = {len(region.cells) * len(region.sides)}"
            )

            to_visit.difference_update(region.cells)
            regions.append(region)

        yield sum(len(region.cells) * len(region.edges) for region in regions)
        yield sum(len(region.cells) * len(region.sides) for region in regions)
