from typing import Any, Callable, Final, Iterator, TypeAlias

import numpy as np
from numpy.typing import NDArray

from ..base import BaseSolver

ImageGrid: TypeAlias = NDArray[np.uint8]


class Grid:
    FREE: Final = 0
    BLOCK: Final = 1
    ROBOT: Final = 2

    robot: tuple[int, int]

    def __init__(self, grid_s: list[str], large: bool):
        grid: list[list[int]] = []
        robot: tuple[int, int] | None = None
        box_counter = 4 if not large else 5
        for i_row, row in enumerate(grid_s):
            row_u: list[int] = []
            for i_col, col in enumerate(row):
                if col in ".@":
                    row_u.extend((Grid.FREE, Grid.FREE) if large else (Grid.FREE,))
                    if col == "@":
                        robot = (i_row, i_col * 2 if large else i_col)
                elif col == "#":
                    row_u.extend((Grid.BLOCK, Grid.BLOCK) if large else (Grid.BLOCK,))
                else:
                    row_u.extend(
                        (box_counter, -box_counter) if large else (box_counter,)
                    )
                    box_counter += 2
            grid.append(row_u)

        self.grid = np.array(grid)

        assert robot is not None
        self.robot = robot

    @property
    def n_rows(self):
        return len(self.grid)

    @property
    def n_columns(self):
        return len(self.grid[0])

    def __len__(self):
        return self.n_rows

    def __iter__(self):
        return iter(self.grid)

    def __getitem__(self, index: tuple[int, int]):
        return self.grid[*index]

    def __setitem__(self, index: tuple[int, int], value: int):
        self.grid[*index] = value

    def is_free(self, row: int, col: int):
        return self[row, col] == Grid.FREE

    def is_block(self, row: int, col: int):
        return self[row, col] == Grid.BLOCK

    def is_box(self, row: int, col: int):
        return (c := self[row, col]) >= 4 and c % 2 == 0

    def is_open_or_close_box(self, row: int, col: int):
        return abs(c := self[row, col]) >= 4 and c % 2 == 1

    def is_open_box(self, row: int, col: int):
        return (c := self[row, col]) >= 4 and c % 2 == 1

    def is_close_box(self, row: int, col: int):
        return self[row, col] < 0

    def _to_char(self, row: int, col: int):
        if self.is_free(row, col):
            return "."
        elif self.is_block(row, col):
            return "#"
        elif self.is_box(row, col):
            return "O"
        elif self.is_open_box(row, col):
            return "["
        else:
            return "]"

    def as_numpy(self):
        arr = self.grid.copy()
        arr[*self.robot] = Grid.ROBOT
        return arr

    def as_printable(self):
        grid_s = [
            [self._to_char(row, col) for col in range(self.n_columns)]
            for row in range(self.n_rows)
        ]
        grid_s[self.robot[0]][self.robot[1]] = "\033[31;1m@\033[00m"
        return "\n".join("".join(row) for row in grid_s)

    def __str__(self):
        return self.as_printable()


class Solver(BaseSolver):
    def save_grid(self, name: str, grid: Grid):
        if self.files:
            self.files.create(name, grid.as_printable().encode(), True)

    def step_part1(self, grid: Grid, move: str):
        match move:
            case "^":
                d_row, d_col = -1, 0
            case ">":
                d_row, d_col = 0, 1
            case "v":
                d_row, d_col = 1, 0
            case "<":
                d_row, d_col = 0, -1
            case _:
                assert False

        row, col = grid.robot
        if grid.is_free(row + d_row, col + d_col):
            grid.robot = (row + d_row, col + d_col)
        elif not grid.is_block(row + d_row, col + d_col):
            n = 1
            while grid.is_box(row + n * d_row, col + n * d_col):
                n += 1

            if grid.is_free(row + n * d_row, col + n * d_col):
                grid.robot = (row + d_row, col + d_col)
                for k in range(2, n + 1):
                    grid[row + k * d_row, col + k * d_col] = grid[
                        row + (k - 1) * d_row, col + (k - 1) * d_col
                    ]
                grid[row + d_row, col + d_col] = Grid.FREE

        return grid

    def step_part2(self, grid: Grid, move: str):
        match move:
            case "^":
                d_row, d_col = -1, 0
            case ">":
                d_row, d_col = 0, 1
            case "v":
                d_row, d_col = 1, 0
            case "<":
                d_row, d_col = 0, -1
            case _:
                assert False

        row, col = grid.robot
        if grid.is_free(row + d_row, col + d_col):
            grid.robot = (row + d_row, col + d_col)
        elif grid.is_block(row + d_row, col + d_col):
            ...
        elif move in "<>":
            n = 1
            while grid.is_open_or_close_box(row, col + n * d_col):
                n += 1

            if grid.is_free(row, col + n * d_col):
                grid.robot = (row, col + d_col)
                for k in range(n, 1, -1):
                    grid[row, col + k * d_col] = grid[row, col + (k - 1) * d_col]
                grid[row + d_row, col + d_col] = Grid.FREE

        elif move in "^v":
            n = 1
            boxes: list[set[int]] = [{col}]
            while True:
                to_move = boxes[-1]
                if any(grid.is_block(row + n * d_row, c) for c in to_move):
                    break
                if all(grid.is_free(row + n * d_row, c) for c in to_move):
                    break

                as_move: set[int] = set()

                for c in to_move:
                    if grid.is_close_box(row + n * d_row, c):
                        as_move.update({c - 1, c})
                    elif grid.is_open_box(row + n * d_row, c):
                        as_move.update({c, c + 1})

                boxes.append(as_move)
                n += 1

            if all(grid.is_free(row + n * d_row, c) for c in boxes[-1]):
                for k, to_move in zip(range(n, 1, -1), boxes[-1:0:-1], strict=True):
                    for c in to_move:
                        grid[row + k * d_row, c] = grid[row + (k - 1) * d_row, c]
                        grid[row + (k - 1) * d_row, c] = Grid.FREE
                grid.robot = (row + d_row, col + d_col)

        return grid

    def run(
        self,
        name: str,
        grid: Grid,
        moves: str,
        fn: Callable[[Grid, str], Grid],
        generate: bool,
    ) -> tuple[Grid, list[ImageGrid]]:
        # initialize
        images: list[ImageGrid] = []

        if generate:
            images.append(grid.as_numpy())

        self.save_grid(f"initial_grid_{name}.txt", grid)

        for move in self.progress.wrap(moves):
            self.logger.debug(f"Move '{move}'...")
            grid = fn(grid, move)

            if generate:
                images.append(grid.as_numpy())

        self.save_grid(f"final_grid_{name}.txt", grid)

        return grid, images

    def solve(self, input: str) -> Iterator[Any]:
        grid_s, moves = input.split("\n\n")
        moves = "".join(moves.split())

        n_boxes = grid_s.count("O")
        colors = np.concatenate(
            [
                np.array(
                    [[255, 255, 255], [64, 64, 64], [255, 0, 0]],
                    dtype=np.uint8,
                ),
                np.random.randint(0, 256, size=(n_boxes, 3), dtype=np.uint8),
            ],
            dtype=np.uint8,
        )

        grid, images = self.run(
            "part1",
            Grid(grid_s.splitlines(), False),
            moves,
            self.step_part1,
            self.files is not None,
        )
        if self.files:
            images = np.stack(images, axis=0)
            images[images >= 2] = 1 + images[images >= 2] // 2
            self.files.video("anim_part1.webm", colors[images])
        yield sum(
            100 * row + col
            for row in range(grid.n_rows)
            for col in range(grid.n_columns)
            if grid.is_box(row, col)
        )

        grid, images = self.run(
            "part2",
            Grid(grid_s.splitlines(), True),
            moves,
            self.step_part2,
            self.files is not None,
        )
        if self.files:
            images = np.abs(np.stack(images, axis=0))
            images[images >= 2] = 1 + images[images >= 2] // 2
            self.files.video("anim_part2.webm", colors[images])
        yield sum(
            100 * row + col
            for row in range(grid.n_rows)
            for col in range(grid.n_columns)
            if grid.is_open_box(row, col)
        )
