from dataclasses import dataclass
from typing import Any, Iterator, Literal, TypeAlias, cast

from ..base import BaseSolver

GateType: TypeAlias = Literal["and", "or", "xor"]


@dataclass(frozen=True, eq=True)
class Gate:
    type: GateType
    lhs: str
    rhs: str

    def __call__(self, lhs: int, rhs: int) -> Any:
        match self.type:
            case "or":
                return int(lhs or rhs)
            case "and":
                return int(lhs and rhs)
            case "xor":
                return int(lhs != rhs)


class Solver(BaseSolver):
    def solve(self, input: str) -> Iterator[Any]:
        inputs_s, gates_s = input.split("\n\n")

        wires: dict[str, int | None] = {}
        for row in inputs_s.splitlines():
            parts = row.split(": ")
            wires[parts[0]] = int(parts[1])

        gates: dict[str, Gate] = {}
        for row in gates_s.splitlines():
            parts = row.split()
            assert parts[4] not in gates
            assert parts[4] not in wires

            lhs, rhs = sorted([parts[0], parts[2]])

            gates[parts[4]] = Gate(cast(GateType, parts[1].lower()), lhs, rhs)
            wires[parts[4]] = None

        if self.files:
            content = "digraph G {\n"
            for wire in wires:
                content += f'{wire} [label="{wire}"]\n'
            for wire, gate in gates.items():
                gate_n = f"G_{wire}"
                content += f'{gate_n} [label="{gate.type.upper()}"]\n'
                content += f"{gate.lhs} -> {gate_n}\n"
                content += f"{gate.rhs} -> {gate_n}\n"
                content += f"{gate_n} -> {wire}\n"
            content += "}\n"
            self.files.create("gates.dot", content.encode(), text=False)

        wires_to_find = set(gates)

        while wires_to_find:
            wires_found: list[str] = []
            for wire in wires_to_find:
                gate = gates[wire]

                lhs, rhs = wires[gate.lhs], wires[gate.rhs]

                if lhs is None or rhs is None:
                    continue

                assert wires[wire] is None
                wires[wire] = gate(lhs, rhs)

                wires_found.append(wire)

            wires_to_find.difference_update(wires_found)

        z_wires = sorted((wire for wire in wires if wire.startswith("z")))
        self.logger.info(
            "binary value is '{}'".format(
                "".join(str(wires[w]) for w in reversed(z_wires))
            )
        )
        yield int("".join(str(wires[w]) for w in reversed(z_wires)), base=2)

        # e00 = x00 ^ y00
        # z00 = e00
        # r00 = x00 & y00

        # e01 = x01 ^ y01
        # z01 = r00 ^ e01
        # a01 = x01 & y01
        # b01 = r00 & e01
        # r01 = a01 | (r00 & e01)

        assert gates["z00"] == Gate("xor", "x00", "y00")

        # normalized names -> gate name
        m_gates: dict[str, str] = {}

        def find_gate(type: GateType, lhs: str, rhs: str):
            try:
                return next(
                    wire
                    for wire, gate in gates.items()
                    if gate.type == type
                    and {gate.lhs, gate.rhs}
                    == {m_gates.get(lhs, lhs), m_gates.get(rhs, rhs)}
                )
            except StopIteration as ex:
                self.logger.info(
                    f"gate {lhs} [{m_gates.get(lhs, lhs)}] {type} {rhs} [{m_gates.get(rhs, rhs)}] not found"
                )
                raise ex

        # find the r00 gate (= x00 & y00)
        m_gates["r00"] = find_gate("and", "x00", "y00")

        swapped: list[str] = []

        for i_wire, z_wire in enumerate(z_wires[1:-1], start=1):
            i2d = f"{i_wire:02d}"
            r0n = f"r{i_wire - 1:02d}"
            m_gates[f"e{i2d}"] = find_gate("xor", f"x{i2d}", f"y{i2d}")

            try:
                z_gate = find_gate("xor", r0n, f"e{i2d}")
            except StopIteration:
                # gate xor not found -> one of the input gate has been swapped
                #
                # assume there is a XOR gate with the remainder, so it is the other
                # input that has been swapped
                assert gates[z_wire].type == "xor"
                assert m_gates.get(r0n, r0n) in (gates[z_wire].lhs, gates[z_wire].rhs)

                wrong_wire_1 = (
                    gates[z_wire].lhs
                    if gates[z_wire].rhs == m_gates.get(r0n, r0n)
                    else gates[z_wire].rhs
                )
                wrong_wire_2 = m_gates[f"e{i2d}"]

                # we are going to fix all the gates (there is probably only 2 but
                # eh...) whose wires needs to be swapped

                self.logger.info(f"swapping {wrong_wire_1} <> {wrong_wire_2}")
                switch = {wrong_wire_1: wrong_wire_2, wrong_wire_2: wrong_wire_1}

                for wire, gate in list(gates.items()):
                    lhs, rhs = (
                        switch.get(gate.lhs, gate.lhs),
                        switch.get(gate.rhs, gate.rhs),
                    )
                    if lhs != gate.lhs or rhs != gate.rhs:
                        gates[wire] = Gate(gate.type, lhs, rhs)

                swapped.extend((wrong_wire_1, wrong_wire_2))

                z_gate = find_gate("xor", r0n, f"e{i2d}")

            if z_gate != z_wire:
                self.logger.info(f"swapping {z_gate} <> {z_wire}")
                gates[z_gate], gates[z_wire] = gates[z_wire], gates[z_gate]
                swapped.extend((z_gate, z_wire))

            m_gates[f"a{i2d}"] = find_gate("and", f"x{i2d}", f"y{i2d}")
            m_gates[f"b{i2d}"] = find_gate("and", r0n, f"e{i2d}")
            m_gates[f"r{i2d}"] = find_gate("or", f"a{i2d}", f"b{i2d}")

        assert len(swapped) == 8
        yield ",".join(sorted(swapped))
